/*
 * Verify that bhyve generates a #GP if any reserved bits in %CR8 are set.
 *
 * kldload ./cr8.ko
 * sysctl debug.cr8.val		# read %CR8
 * sysctl debug.cr8.val=0	# set %CR8 to 0
 * sysctl debug.cr8.val=0x10	# GPF on attempt to set reserved bit in %CR8
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/sysctl.h>
#include <sys/module.h>

static u_long
rcr8(void)
{
	u_long	data;

	__asm __volatile("movq %%cr8,%0" : "=r" (data));
	return (data);
}

static __inline void
load_cr8(u_long data)
{

	__asm __volatile("movq %0,%%cr8" : : "r" (data));
}

static int
sysctl_cr8_val(SYSCTL_HANDLER_ARGS)
{
	int error;
	u_long val;

	val = rcr8();
	error = sysctl_handle_long(oidp, &val, 0, req);
	if (error || req->newptr == NULL)
		return (error);

	load_cr8(val);
	return (0);
}

SYSCTL_NODE(_debug, OID_AUTO, cr8, CTLFLAG_RW, 0, "");

SYSCTL_PROC(_debug_cr8, OID_AUTO, val, CTLTYPE_ULONG|CTLFLAG_RW, 0, 0, 
    sysctl_cr8_val, "LU", "");

static int
cr8_modevent(module_t mod, int type, void *unused)
{

        switch (type) {
        case MOD_LOAD:
		break;
	case MOD_UNLOAD:
		break;
	default:
		return (EINVAL);
	}

	return (0);
}

static moduledata_t cr8_mod_data= {
	"cr8",
	cr8_modevent,
};

DECLARE_MODULE(test, cr8_mod_data, SI_SUB_SMP + 1, SI_ORDER_ANY);
